@@ -26,17 +26,19 @@
 #define need_PLparser()  ObjectPad__need_PLparser(aTHX)
 void ObjectPad__need_PLparser(pTHX); /* in Object/Pad.xs */
 
-FieldMeta *ObjectPad_mop_create_field(pTHX_ SV *fieldname, ClassMeta *classmeta)
+FieldMeta *ObjectPad_mop_create_field(pTHX_ SV *fieldname, FIELDOFFSET fieldix, ClassMeta *classmeta)
 {
   FieldMeta *fieldmeta;
   Newx(fieldmeta, 1, FieldMeta);
 
-  assert(classmeta->next_fieldix > -1);
+  assert(fieldix > -1);
 
   *fieldmeta = (FieldMeta){
-    .name    = SvREFCNT_inc(fieldname),
-    .class   = classmeta,
-    .fieldix = classmeta->next_fieldix,
+    LINNET_INIT(LINNET_VAL_FIELDMETA)
+    .name      = SvREFCNT_inc(fieldname),
+    .is_direct = true,
+    .class     = classmeta,
+    .fieldix   = fieldix,
   };
 
   return fieldmeta;
@@ -66,6 +68,7 @@ static void S_mop_field_set_param(pTHX_ FieldMeta *fieldmeta, SV *paramname)
   Newx(parammeta, 1, struct ParamMeta);
 
   *parammeta = (struct ParamMeta){
+    LINNET_INIT(LINNET_VAL_PARAMMETA)
     .name  = SvREFCNT_inc(paramname),
     .class = classmeta,
     .type  = PARAM_FIELD,
@@ -314,7 +317,7 @@ SV *ObjectPad_get_obj_fieldsv(pTHX_ SV *self, ClassMeta *classmeta, FieldMeta *f
     if(!svp)
       croak("Cannot fetch role field value from a non-applied instance");
 
-    RoleEmbedding *embedding = (RoleEmbedding *)*svp;
+    RoleEmbedding *embedding = MUST_ROLEEMBEDDING(*svp);
 
     fieldstore = get_obj_fieldstore(self, embedding->classmeta->repr, true);
     fieldix = fieldmeta->fieldix + embedding->offset;
@@ -826,6 +829,25 @@ static struct FieldHookFuncs fieldhooks_accessor = {
   .seal  = &fieldhook_accessor_seal,
 };
 
+/* :inheritable */
+
+static bool fieldhook_inheritble_apply(pTHX_ FieldMeta *fieldmeta, SV *value, SV **attrdata_ptr, void *_funcdata)
+{
+  HV *hints = GvHV(PL_hintgv);
+  if(!hv_fetchs(hints, "Object::Pad/experimental(inherit_field)", 0))
+    Perl_ck_warner(aTHX_ packWARN(WARN_EXPERIMENTAL),
+      "inheriting fields is experimental and may be changed or removed without notice");
+
+  fieldmeta->is_inheritable = true;
+
+  return false;
+}
+
+static struct FieldHookFuncs fieldhooks_inheritable = {
+  .ver   = OBJECTPAD_ABIVERSION,
+  .apply = &fieldhook_inheritble_apply,
+};
+
 void ObjectPad_register_field_attribute(pTHX_ const char *name, const struct FieldHookFuncs *funcs, void *funcdata)
 {
   if(funcs->ver < 57)
@@ -859,4 +881,7 @@ void ObjectPad__boot_fields(pTHX)
   register_field_attribute("writer",   &fieldhooks_writer,   NULL);
   register_field_attribute("mutator",  &fieldhooks_mutator,  NULL);
   register_field_attribute("accessor", &fieldhooks_accessor, NULL);
+
+  // TODO: temporary name
+  register_field_attribute("inheritable", &fieldhooks_inheritable, NULL);
 }
